/*
Copyright 2017 Heptio Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package client

import (
	"os"

	"github.com/spf13/pflag"

	"github.com/heptio/ark/pkg/generated/clientset"
)

// Factory knows how to create an ArkClient.
type Factory interface {
	// BindFlags binds common flags such as --kubeconfig to the passed-in FlagSet.
	BindFlags(flags *pflag.FlagSet)
	// Client returns an ArkClient. It uses the following priority to specify the cluster
	// configuration:  --kubeconfig flag, KUBECONFIG environment variable, in-cluster configuration.
	Client() (clientset.Interface, error)
}

type factory struct {
	flags      *pflag.FlagSet
	kubeconfig string
}

// NewFactory returns a Factory.
func NewFactory() Factory {
	f := &factory{
		flags: pflag.NewFlagSet("", pflag.ContinueOnError),
	}
	f.flags.StringVar(&f.kubeconfig, "kubeconfig", "", "Path to the kubeconfig file to use to talk to the Kubernetes apiserver. If unset, try the environment variable KUBECONFIG, as well as in-cluster configuration")

	return f
}

func (f *factory) BindFlags(flags *pflag.FlagSet) {
	flags.AddFlagSet(f.flags)
}

func (f *factory) Client() (clientset.Interface, error) {
	kubeconfig := f.kubeconfig
	if kubeconfig == "" {
		// if the command line flag was not specified, try the environment variable
		kubeconfig = os.Getenv("KUBECONFIG")
	}

	clientConfig, err := Config(kubeconfig)
	if err != nil {
		return nil, err
	}

	arkClient, err := clientset.NewForConfig(clientConfig)
	if err != nil {
		return nil, err
	}
	return arkClient, nil
}
